/***********************************************************************
 @C-file{
    author              = "Nelson H. F. Beebe",
    version             = "2.03",
    date                = "28 January 2004",
    time                = "09:30:09 MST",
    filename            = "bsplit.c",
    address             = "Center for Scientific Computing
                           University of Utah
                           Department of Mathematics, 322 INSCC
                           155 S 1400 E RM 233
                           Salt Lake City, UT 84112-0090
                           USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 585 1640, +1 801 581 4148",
    URL                 = "http://www.math.utah.edu/~beebe",
    checksum            = "35158 422 1371 10057",
    email               = "beebe@math.utah.edu, beebe@acm.org,
                           beebe@computer.org (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "binary file split",
    supported           = "yes",
    docstring           = "Split one or more binary file into parts,
                           usually for packaging on floppy disks, or
                           for better restartability of file copies
                           over poor network connections.

                           The checksum field above contains a CRC-16
                           checksum as the first value, followed by the
                           equivalent of the standard UNIX wc (word
                           count) utility output of lines, words, and
                           characters.  This is produced by Robert
                           Solovay's checksum utility.",
 }
***********************************************************************/

#include "config.h"
#include "stdc.h"
#include "xctype.h"
#include "xstdio.h"
#include "xstdlib.h"
#include "xstring.h"

#ifdef __COMPILER_KCC__
#define BINREAD		"rb8"
#define BINWRITE	"wb8"
#else /* not __COMPILER_KCC__ */
#ifdef VMS
#define BINREAD		"rb","rfm=fix","bls=512","mrs=512"
#define BINWRITE	"wb","rfm=fix","bls=512","mrs=512"
#else /* assume Unix style */
#define BINREAD		"rb"
#define BINWRITE	"wb"
#endif
#endif /* __COMPILER_KCC__ */

#define DEFAULT_SIZE	(1423L * 1024L)	/* just below capacity of IBM PC 3.5in floppy */
#define MIN_SIZE	(1024L)		/* no pieces are ever smaller than this */

#if !defined(_MAX_PATH)
#if defined(OS_PCDOS)
#define _MAX_PATH 260			/* Microsoft C 6.0 and 7.0 use this value */
#else
#define _MAX_PATH 1024			/* big enough for all current file systems */
#endif
#endif

#if defined(OS_PCDOS) || defined(OS_VMS)
#define BSPLIT_FORMAT "%.256s.%03d"	/* 256 + 1 + 3 = 260 (== _MAX_PATH) */
#define is_option_char(c)	(((c) == '/') || ((c) == '-'))
#else
#define BSPLIT_FORMAT "%.1020s-%03d"
#define is_option_char(c)	((c) == '-')
#endif

int			main ARGS ((int argc, char *argv[]));
static const char 	*base_name ARGS ((const char *s));
static void		bsplit ARGS ((long size, const char *name));
static long		bytes ARGS ((const char *s));
static void		help ARGS ((const char *program));
static void		help_and_stop ARGS ((const char *program));
static void		usage ARGS ((const char *program));
static void		usage_and_die ARGS ((const char *program));
static void		usage_and_stop ARGS ((const char *program));
static void		version ARGS ((const char *program));
static void		version_and_stop ARGS ((const char *program));


#if STDC
int
main (int argc, char *argv[])
#else
int
main (argc, argv)
int argc;
char *argv[];
#endif
{
    long size;
    size_t k;
    size_t nfiles;
    const char *option;

    size = DEFAULT_SIZE;

    nfiles = 0;

    for (k = 1; k < (size_t)argc; ++k)
    {
	option = (const char *) argv[k];
	if (is_option_char (option[0]))
	{
	    if ((option[0] == '-') && (option[1] == '-'))
		option++;	/* handle GNU- and POSIX-style --option */

	    switch (option[1])
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':		/* have -nnn */
		size = bytes (&option[1]);
		break;

	    case '?':		/* -? */
	    case 'h':		/* -h */
	    case 'H':		/* -H */
		help_and_stop (argv[0]);
		break;		/* NOT REACHED */

	    case 'b':		/* expect -b nnn */
	    case 'B':
		if (++k >= (size_t)argc)
		    break;	/* missing size argument */
		else
		    size = bytes (argv[k]);
		break;

	    case 'v':		/* have -v */
	    case 'V':
		version_and_stop (argv[0]);
		break;		/* NOT REACHED */

	    default:		/* unrecognized option */
		(void) fprintf (stderr,"unrecognized option: [%s]\n", argv[k]);
		usage_and_die (argv[0]);
		break;		/* NOT REACHED */
	    }
	}
	else
	    /* expect filename */
	{
	    bsplit (size, argv[k]);
	    nfiles++;
	}
    }

    if (nfiles == 0)		/* then split input from stdin */
	bsplit (size, (const char *) NULL);

    exit (EXIT_SUCCESS);
    return (EXIT_SUCCESS);	/* NOT REACHED */
}


#if STDC
static const char *
base_name (const char *s)
#else
static const char *
base_name (s)
const char *s;
#endif
{				/* return a pointer to a (possibly truncated) copy of s[] */

#if defined(OS_PCDOS)
    static char outname[_MAX_PATH];
    char *p;

    /* Because of the 8+3 file name rule of IBM PC DOS, we must
       truncate the name at the file extension, if any, so that we
       supply a numeric extension .nnn later: see BSPLIT_FORMAT
       above. */
    (void) strncpy (outname, s, sizeof (outname));
    outname[sizeof (outname) - 1] = '\0';	/* ensure NUL termination */

    for (p = strchr (outname, '\0') - 1; p >= &outname[0]; --p)
    {
	if ((*p == '/') || (*p == '\\'))	/* reached directory separator */
	    break;		/* so no period found */
	else if (*p == '.')	/* reached period */
	{
	    *p = '\0';		/* so delete it */
	    break;
	}
    }
    return ((const char *) &outname[0]);
#else
    return (s);
#endif

}


#if STDC
static void
bsplit (long size, const char *inname)
#else
static void
bsplit (size, inname)
long size;
const char *inname;
#endif
{
    register FILE *fpin;
    register FILE *fpout;
    static char outname[_MAX_PATH];
    register long n;
    long npart;
    register int c;

    fpin = (inname == (const char *) NULL) ? stdin : fopen (inname, BINREAD);
    if (fpin == (FILE *) NULL)
    {
	(void) fprintf (stderr, "?Cannot open file [%s]\n", inname);
	return;
    }
    (void) fprintf (stderr, "[%s]\n",
		    ((inname == (const char *) NULL) ? "stdin" : inname));
    for (npart = 1; !feof (fpin); ++npart)
    {
	(void) sprintf (outname, BSPLIT_FORMAT,
	   ((inname == (const char *) NULL) ? "stdout" : base_name (inname)),
			(int)npart);
	fpout = fopen (outname, BINWRITE);
	if (fpout == (FILE *) NULL)
	{
	    (void) fprintf (stderr,
			    "?Cannot open file [%s]\n", outname);
	    return;
	}
	(void) fprintf (stderr, "\t[%s]", outname);
	for (n = 0; n < size; ++n)
	{
	    c = getc (fpin);
	    if ((c == EOF) && feof (fpin))
		break;
	    putc (c, fpout);
	}
	(void) fclose (fpout);
	(void) fprintf (stderr, " %d bytes [OK]\n", (int)n);
    }
}


#if STDC
static long
bytes (const char *s)
#else
static long
bytes (s)
const char *s;
#endif
{
    const char *p;
    long size;

    size = atol (s);

    if (size < 0)
	size = -size;		/* ignore sign */

    p = strchr (s, '\0');

    if (p == (const char *) NULL)	/* cannot happen */
	return (0);

    switch (islower ((int) p[-1]) ? toupper ((int) p[-1]) : (int) p[-1])
    {				/* handle K M G T P E suffixes */
    case 'E':			/* exabytes */
	size *= 1024L;		/* fall through  */

    case 'P':			/* petabytes */
	size *= 1024L;		/* fall through  */

    case 'T':			/* terabytes */
	size *= 1024L;		/* fall through  */

    case 'G':			/* gigabytes */
	size *= 1024L;		/* fall through  */

    case 'M':			/* megabytes */
	size *= 1024L;		/* fall through  */

    case 'K':			/* kilobytes */
	size *= 1024L;		/* fall through  */

    default:
	/* NO-OP */ ;
    }

    size = 512L * ((size + 511L) / 512L);	/* ensure multiple of 512 bytes */

    if (size < MIN_SIZE)	/* guarantee a reasonable size */
    {
	size = DEFAULT_SIZE;
	(void) fprintf (stderr,
	      "requested size %s too small: reverting to default of %ldK\n",
			s, size / 1024L);
    }

    return (size);
}


#if STDC
static void
help (const char *program)
#else
static void
help (program)
const char *program;
#endif
{
    usage (program);
    (void) fprintf (stderr, "\nThe options are:\n");
    (void) fprintf (stderr, "\t-?                      Display brief help message and quit\n");
    (void) fprintf (stderr, "\t-b nnn[K|M|G|T|P|E]     Set split size in bytes\n");
    (void) fprintf (stderr, "\t-h                      Display brief help message and quit\n");
    (void) fprintf (stderr, "\t-nnn[K|M|G|T|P|E]       Set split size in bytes\n");
    (void) fprintf (stderr, "\t-v                      Display version number and quit\n");
}


#if STDC
static void
help_and_stop (const char *program)
#else
static void
help_and_stop (program)
const char *program;
#endif
{
    help (program);
    exit (EXIT_SUCCESS);
}


#if STDC
static void
usage (const char *program)
#else
static void
usage (program)
const char *program;
#endif
{
    (void) fprintf (stderr,
	    "Usage: %s [-?] [[-b |-]bytes[K|M|G|T|P|E]] [-h] [-v] filename(s)\n",
		    program);
}


#if STDC
static void
usage_and_die (const char *program)
#else
static void
usage_and_die (program)
const char *program;
#endif
{
    usage (program);
    exit (EXIT_FAILURE);
}


#if STDC
static void
usage_and_stop (const char *program)
#else
static void
usage_and_stop (program)
const char *program;
#endif
{
    usage (program);
    exit (EXIT_SUCCESS);
}


#if STDC
static void
version (const char *program)
#else
static void
version (program)
const char *program;
#endif
{
    (void) fprintf (stderr, "%s Version %s %s\n", program, VERSION, DATE);
}


#if STDC
static void
version_and_stop (const char *program)
#else
static void
version_and_stop (program)
const char *program;
#endif
{
    version (program);
    exit (EXIT_SUCCESS);
}
